package com.example.pmulabs.api.mediator

import androidx.paging.ExperimentalPagingApi
import androidx.paging.LoadType
import androidx.paging.PagingState
import androidx.paging.RemoteMediator
import androidx.room.withTransaction
import com.example.pmulabs.api.NewsPortalService
import com.example.pmulabs.api.model.toTag
import com.example.pmulabs.room.database.NewsPortalDatabase
import com.example.pmulabs.room.models.RemoteKeyType
import com.example.pmulabs.room.models.RemoteKeys
import com.example.pmulabs.room.models.Tag
import com.example.pmulabs.room.repository.OfflineRemoteKeyRepository
import com.example.pmulabs.room.repository.OfflineTagRepository
import retrofit2.HttpException
import java.io.IOException

@OptIn(ExperimentalPagingApi::class)
class TagRemoteMediator(
    private val service: NewsPortalService,
    private val dbTagRepository: OfflineTagRepository,
    private val dbRemoteKeyRepository: OfflineRemoteKeyRepository,
    private val database: NewsPortalDatabase
) : RemoteMediator<Int, Tag>() {

    override suspend fun initialize(): InitializeAction {
        return InitializeAction.LAUNCH_INITIAL_REFRESH
    }

    override suspend fun load(
        loadType: LoadType,
        state: PagingState<Int, Tag>
    ): MediatorResult {
        val page = when (loadType) {
            LoadType.REFRESH -> {
                val remoteKeys = getRemoteKeyClosestToCurrentPosition(state)
                remoteKeys?.nextKey?.minus(1) ?: 1
            }

            LoadType.PREPEND -> {
                val remoteKeys = getRemoteKeyForFirstItem(state)
                remoteKeys?.prevKey
                    ?: return MediatorResult.Success(endOfPaginationReached = remoteKeys != null)
            }

            LoadType.APPEND -> {
                val remoteKeys = getRemoteKeyForLastItem(state)
                remoteKeys?.nextKey
                    ?: return MediatorResult.Success(endOfPaginationReached = remoteKeys != null)
            }
        }

        try {
            val tags = service.getTags(page, state.config.pageSize).map { it.toTag() }
            val endOfPaginationReached = tags.isEmpty()
            database.withTransaction {
                if (loadType == LoadType.REFRESH) {
                    dbRemoteKeyRepository.deleteRemoteKey(RemoteKeyType.TAG)
                    dbTagRepository.clearTags()
                }
                val prevKey = if (page == 1) null else page - 1
                val nextKey = if (endOfPaginationReached) null else page + 1
                val keys = tags.map {
                    it.id?.let { it1 ->
                        RemoteKeys(
                            entityId = it1,
                            type = RemoteKeyType.TAG,
                            prevKey = prevKey,
                            nextKey = nextKey
                        )
                    }
                }
                dbRemoteKeyRepository.createRemoteKeys(keys)
                dbTagRepository.insertTags(tags)
            }
            return MediatorResult.Success(endOfPaginationReached = endOfPaginationReached)
        } catch (exception: IOException) {
            return MediatorResult.Error(exception)
        } catch (exception: HttpException) {
            return MediatorResult.Error(exception)
        }
    }

    private suspend fun getRemoteKeyForLastItem(state: PagingState<Int, Tag>): RemoteKeys? {
        return state.pages.lastOrNull { it.data.isNotEmpty() }?.data?.lastOrNull()
            ?.let { tag ->
                tag.id?.let { dbRemoteKeyRepository.getAllRemoteKeys(it, RemoteKeyType.TAG) }
            }
    }

    private suspend fun getRemoteKeyForFirstItem(state: PagingState<Int, Tag>): RemoteKeys? {
        return state.pages.firstOrNull { it.data.isNotEmpty() }?.data?.firstOrNull()
            ?.let { tag ->
                tag.id?.let { dbRemoteKeyRepository.getAllRemoteKeys(it, RemoteKeyType.TAG) }
            }
    }

    private suspend fun getRemoteKeyClosestToCurrentPosition(
        state: PagingState<Int, Tag>
    ): RemoteKeys? {
        return state.anchorPosition?.let { position ->
            state.closestItemToPosition(position)?.id?.let { tagid ->
                dbRemoteKeyRepository.getAllRemoteKeys(tagid, RemoteKeyType.TAG)
            }
        }
    }

}