package com.example.pmulabs.api.repository

import androidx.paging.ExperimentalPagingApi
import androidx.paging.Pager
import androidx.paging.PagingConfig
import androidx.paging.PagingData
import com.example.pmulabs.api.NewsPortalService
import com.example.pmulabs.api.mediator.ArticleRemoteMediator
import com.example.pmulabs.api.model.toArticle
import com.example.pmulabs.api.model.toArticleRemote
import com.example.pmulabs.room.AppContainer
import com.example.pmulabs.room.database.NewsPortalDatabase
import com.example.pmulabs.room.models.Article
import com.example.pmulabs.room.repository.ArticleRepository
import com.example.pmulabs.room.repository.OfflineArticleRepository
import com.example.pmulabs.room.repository.OfflineRemoteKeyRepository
import kotlinx.coroutines.flow.Flow

class RestArticleRepository(
    private val service: NewsPortalService,
    private val dbArticleRepository: OfflineArticleRepository,
    private val dbRemoteKeyRepository: OfflineRemoteKeyRepository,
    private val dbTagRepository: RestTagRepository,
    private val database: NewsPortalDatabase
) : ArticleRepository {
    override suspend fun insertArticle(article: Article) {
        service.createArticle(article.toArticleRemote()).toArticle()
    }

    override suspend fun updateArticle(article: Article) {
        article.id?.let { service.updateArticle(it, article.toArticleRemote()).toArticle() }
    }

    override suspend fun deleteArticle(article: Article) {
        article.id?.let { service.deleteArticle(it).toArticle() }
    }

    override suspend fun getAllArticles(): List<Article> {
        dbTagRepository.getAllTags()
        val existArticles = dbArticleRepository.getAllArticles().associateBy { it.id }.toMutableMap()

        service.getAllArticles()
            .map { it.toArticle() }
            .forEach { art ->
                val existArt = existArticles[art.id]
                if (existArt == null) {
                    dbArticleRepository.insertArticle(art)
                } else if (existArt != art) {
                    dbArticleRepository.updateArticle(art)
                }
                existArticles[art.id] = art
            }

        return existArticles.map { it.value }.sortedBy { it.id }
    }

    override suspend fun getArticleById(idArticle: Int): Article?  =
        idArticle?.let { service.getArticle(it).toArticle() }

    override fun getArticles(): Flow<PagingData<Article>> {
        val pagingSourceFactory = { dbArticleRepository.getAllArticlesPagingSource() }

        @OptIn(ExperimentalPagingApi::class)
        return Pager(
            config = PagingConfig(
                pageSize = AppContainer.LIMIT,
                enablePlaceholders = false
            ),
            remoteMediator = ArticleRemoteMediator(
                service,
                dbArticleRepository,
                dbRemoteKeyRepository,
                dbTagRepository,
                database,
            ),
            pagingSourceFactory = pagingSourceFactory
        ).flow
    }
}