package com.example.pmulabs.api.repository

import androidx.paging.ExperimentalPagingApi
import androidx.paging.Pager
import androidx.paging.PagingConfig
import androidx.paging.PagingData
import com.example.pmulabs.api.NewsPortalService
import com.example.pmulabs.api.mediator.CommentRemoteMediator
import com.example.pmulabs.api.model.toComment
import com.example.pmulabs.api.model.toCommentRemote
import com.example.pmulabs.api.model.toUser
import com.example.pmulabs.room.AppContainer
import com.example.pmulabs.room.database.NewsPortalDatabase
import com.example.pmulabs.room.models.Comment
import com.example.pmulabs.room.models.User
import com.example.pmulabs.room.repository.CommentRepository
import com.example.pmulabs.room.repository.OfflineCommentRepository
import com.example.pmulabs.room.repository.OfflineRemoteKeyRepository
import kotlinx.coroutines.flow.Flow

class RestCommentRepository(
    private val service: NewsPortalService,
    private val dbCommentRepository: OfflineCommentRepository,
    private val dbRemoteKeyRepository: OfflineRemoteKeyRepository,
    private val dbArticleRepository: RestArticleRepository,
    private val database: NewsPortalDatabase
) : CommentRepository {
    override suspend fun insertComment(comment: Comment) {
        service.createComment(comment.toCommentRemote()).toComment()
    }

    override suspend fun updateComment(comment: Comment) {
        comment.id?.let { service.updateComment(it, comment.toCommentRemote()).toComment() }
    }

    override suspend fun deleteComment(comment: Comment) {
        comment.id?.let { service.deleteComment(it).toComment() }
    }

    override suspend fun getAllComments(): List<Comment> {
        val existComms = dbCommentRepository.getAllComments().associateBy { it.id }.toMutableMap()

        service.getAllComments()
            .map { it.toComment() }
            .forEach { comm ->
                val existComm = existComms[comm.id]
                if (existComm == null) {
                    dbCommentRepository.insertComment(comm)
                } else if (existComm != comm) {
                    dbCommentRepository.updateComment(comm)
                }
                existComms[comm.id] = comm
            }

        return existComms.map { it.value }.sortedBy { it.id }
    }

    override suspend fun getCountComment(idArticle: Int?): Int? {
        return idArticle?.let { service.getCountComment(it).size }
    }

    override suspend fun getUser(commId : Int) : User? =
        commId?.let { service.getUser(it).toUser() }


    override fun getComments(): Flow<PagingData<Comment>> {
        val pagingSourceFactory = { dbCommentRepository.getAllCommentsPagingSource() }

        @OptIn(ExperimentalPagingApi::class)
        return Pager(
            config = PagingConfig(
                pageSize = AppContainer.LIMIT,
                enablePlaceholders = false
            ),
            remoteMediator = CommentRemoteMediator(
                service,
                dbCommentRepository,
                dbRemoteKeyRepository,
                dbArticleRepository,
                database,
            ),
            pagingSourceFactory = pagingSourceFactory
        ).flow
    }
}