package com.example.pmulabs.api.repository

import androidx.paging.ExperimentalPagingApi
import androidx.paging.Pager
import androidx.paging.PagingConfig
import androidx.paging.PagingData
import com.example.pmulabs.api.NewsPortalService
import com.example.pmulabs.api.mediator.TagRemoteMediator
import com.example.pmulabs.api.model.toTag
import com.example.pmulabs.api.model.toTagRemote
import com.example.pmulabs.room.AppContainer
import com.example.pmulabs.room.database.NewsPortalDatabase
import com.example.pmulabs.room.models.Tag
import com.example.pmulabs.room.repository.OfflineRemoteKeyRepository
import com.example.pmulabs.room.repository.OfflineTagRepository
import com.example.pmulabs.room.repository.TagRepository
import kotlinx.coroutines.flow.Flow

class RestTagRepository(
    private val service: NewsPortalService,
    private val dbTagRepository: OfflineTagRepository,
    private val dbRemoteKeyRepository: OfflineRemoteKeyRepository,
    private val database: NewsPortalDatabase
) : TagRepository {
    override suspend fun insertTag(tag: Tag) {
        service.createTag(tag.toTagRemote()).toTag()
    }

    override suspend fun updateTag(tag: Tag) {
        tag.id?.let { service.updateTag(it, tag.toTagRemote()).toTag() }
    }

    override suspend fun deleteTag(tag: Tag) {
        tag.id?.let { service.deleteTag(it).toTag() }
    }

    override suspend fun getAllTags(): List<Tag> {
        val existTags = dbTagRepository.getAllTags().associateBy { it.id }.toMutableMap()

        service.getAllTags()
            .map { it.toTag() }
            .forEach { tag ->
                val existTag = existTags[tag.id]
                if (existTag == null) {
                    dbTagRepository.insertTag(tag)
                } else if (existTag != tag) {
                    dbTagRepository.updateTag(tag)
                }
                existTags[tag.id] = tag
            }

        return existTags.map { it.value }.sortedBy { it.id }
    }

    override suspend fun getTagById(idTag: Int): Tag? =
        idTag?.let { service.getTag(it).toTag() }

    override suspend fun getTagByName(nameTag: String): Tag? =
        nameTag?.let { service.getTagByName(it)[0].toTag() }

    override fun getTags(): Flow<PagingData<Tag>> {
        val pagingSourceFactory = { dbTagRepository.getAllTagsPagingSource() }

        @OptIn(ExperimentalPagingApi::class)
        return Pager(
            config = PagingConfig(
                pageSize = AppContainer.LIMIT,
                enablePlaceholders = false
            ),
            remoteMediator = TagRemoteMediator(
                service,
                dbTagRepository,
                dbRemoteKeyRepository,
                database,
            ),
            pagingSourceFactory = pagingSourceFactory
        ).flow
    }

}